#common subs for CCM and SCM

package CCM::Common;

use Exporter;

our @ISA = ('Exporter');
our @EXPORT = qw(extract_using_region_info
		 extract_using_region_info_list
		 extract_region_by_boundary_info
		 get_boundary_info
		 extract_non_seed_page);



use strict;
use warnings;
use diagnostics;
use HTML::TreeBuilder;
use CCM::Util;
use CCM::Debug;
use CCM::PerlUtil;

=head1 NAME

Common

=head1 SYNOPSIS

Common functions that can be shared by CCM and SCM.

=head1 DESCRIPTION

No matter how we get region info (data region node and neighboring level_1 nodes), the remaining operations are the same and listed here.

=cut

sub get_boundary_info($\@\%);
sub extract_region_by_boundary_info($$\@\%$);

sub extract_non_seed_page($$\@\%) {
  my ($filepath, $keyword, $phrase_list_ref, $boundary_info_list_ref) = @_;
  my $tree = get_html_tree($filepath);
  my @phrase_list = @$phrase_list_ref;
  my %boundary_info_list = %$boundary_info_list_ref;
  if ($#phrase_list < 1) {
    print "Cannot extract page that has less than two occurrances of keywords: $keyword.\n";
    return;
  }
  my @records_list = ();
  my %phrase_hash = get_phrase_hash(@phrase_list);
  foreach my $region_address (keys %boundary_info_list) {
    my $region = $tree->address($region_address);
    next unless defined($region);
    my @boundary_info = @{$boundary_info_list{$region_address}};
    my ($records_ref, $matched_phrases_ref) = 
    extract_region_by_boundary_info($region,$region_address,  
				    @boundary_info,
				    %phrase_hash, 
				    $keyword);
    my @records = @$records_ref;
    my @matched_phrases = @$matched_phrases_ref;
    foreach my $phrase (@matched_phrases) {
      $phrase_hash{$phrase} -= 1;
    }
    push @records_list, \@records;
  }
  return ($tree, \@records_list);
}

sub extract_using_region_info(\@\%$) {
  #input: * data region info: ($region_node, $n1, $n2, \@seed_lineage)
  #       * page phrase hash
  #       * keyword
  #output: * list of "record nodes"
  #        * boundary info list
  my ($region_info_ref, $phrase_hash_ref, $keyword) = @_;
  my @region_info = @$region_info_ref;
  my %unmatched_phrase_hash = %$phrase_hash_ref;
  my $region = $region_info[0];
  my @n1_n2 = @region_info[1..2];
  my $seed_lineage_ref = $region_info[3];
  my @boundary_info = get_boundary_info($region, @n1_n2, %unmatched_phrase_hash);
  push @boundary_info, $seed_lineage_ref;
  my ($records_ref, $matched_phrases_ref) = 
    extract_region_by_boundary_info($region,$region->address,  
				    @boundary_info,
				    %$phrase_hash_ref, 
				    $keyword);
  my @records = @$records_ref;
  my @matched_phrases = @$matched_phrases_ref;
  foreach my $phrase (@matched_phrases) {
    $unmatched_phrase_hash{$phrase} -= 1;
  }
  return (\@records, \@boundary_info, \%unmatched_phrase_hash);
}

sub extract_using_region_info_list (\@\%$){
  #input: * list of data region info: ($region_node, $n1, $n2)
  #       * page phrase hash
  #       * keyword
  #output: * list of "record nodes"
  #        * boundary info list
  my ($region_info_list_ref, $phrase_hash_ref, $keyword) = @_;
  my @record_list = ();
  my %boundary_info_list = ();
  my %unmatched_phrase_hash = %$phrase_hash_ref;
  my $unmatched_phrase_hash_ref = \%unmatched_phrase_hash;
  foreach my $region_info_ref(@$region_info_list_ref) {
    my ($records_ref, $boundary_info_ref, $unmatched_phrase_hash_ref) = 
      extract_using_region_info(@$region_info_ref, %$unmatched_phrase_hash_ref,
				$keyword);
     push @record_list, $records_ref;
     my $region = ${$region_info_ref}[0];
     $boundary_info_list{$region->address} = $boundary_info_ref;
  }
  my @unmatched_phrases = ();
  while (my ($key, $value) = each(%unmatched_phrase_hash)) {
    if ($unmatched_phrase_hash{$key}) {
      push @unmatched_phrases, $key;
    }
  }
  return (\@record_list, \%boundary_info_list, \@unmatched_phrases);
}

sub extract_region_by_boundary_info($$\@\%$){
  #input: *node to process, could be root or region node
  #       *region_node address
  #       *boundary_info (keyword_ele_index, taglist, \@seed_lineage)
  #       *page phrase hash
  #       *keyword
  #output: *record list
  #        *matched phrase list
  #trace("extract_region_by_boundary_info function");
  my ($tree, $region_address, $boundary_ref, 
      $phrase_ref, $keyword) = @_;
  my $data_region = $tree->address($region_address);
  my @boundary_info = @$boundary_ref;
  my $keyword_ele_index = $boundary_info[0];
  my @boundary_tag_list = @{$boundary_info[1]};
  #trace "boundary tags: $#boundary_tag_list";
  #map {print $_ . " "} @boundary_tag_list;
  my @seed_lineage =  @{$boundary_info[2]};
  my $seed_relative_addr = $seed_lineage[0];
  my $seed_tag = $seed_lineage[1];
  my %phrase_hash = %$phrase_ref;
  my @children = $data_region->content_list();
  #print $data_region->as_HTML . "\n";
  my $child = get_first_child_element($data_region);

  #use the boundary info to extract current region
  my @record_list = ();
  while (defined($child)) {
#    trace("child address: " . $child->address . "\n");
    my $test_ele = $child;
    my @record = ();
    my $has_record = 0;
    my $idx = 0; #try to match keyword_ele_index
TAGMATCH: for my $boundary_tag (@boundary_tag_list) {
      if ($child->tag ne $boundary_tag) {
	$has_record = 0;
	last;
      } else { #one tag match
	# we want to validate that this is indeed a record
	if ($idx == $keyword_ele_index) {
	  #trace("seed relative: $seed_relative_addr");
	  my $test_address = $child->address . "." . $seed_relative_addr;
	  #trace ("test_address: $test_address");
	  my $test_ele = $data_region->address($test_address);
	 
	  if (!defined($test_ele) 
	      || (ref $test_ele) ne "HTML::Element"
	      || ($test_ele->tag ne $seed_tag)) {
	    $has_record = 0;
	    last TAGMATCH;
	  }
	  $has_record = 1; 
	}
	push @record, $child;
	$child = get_right_element($child);
      }
      ++$idx;
    }
    
    if ($has_record) {
      push @record_list, [@record];
    } elsif (defined($child)) {
      $child = get_right_element($child);
    }
  }
#  trace("after loop:\n\n");
  for my $each_record (@record_list) {
    my @record = @$each_record;
    my $ele = $record[$keyword_ele_index];
    my @phrases = get_phrase_list($ele->as_HTML, $keyword);
    for my $phrase (@phrases) {
      $phrase_hash{$phrase} = 0;
    }
  }
  #output_record_list(@record_list);
  my @matched_phrase_list = ();
  while(my ($key, $value) = each(%phrase_hash)) {
     if (!$phrase_hash{$key}) {
      push @matched_phrase_list, $key;
    }
  }
  return (\@record_list, \@matched_phrase_list);
}

sub get_boundary_info($\@\%) {
  #input: *region node
  #       *n1 and n2
  #       *phrase hash
  #output:*boundary info ($keyword_ele_index, \@boundary_tag_list, 
  #                       \@lineage_list)

  #trace("get_boundary_info sub...");
  my ($data_region, $n_ref, $phrase_hash_ref) = @_;
  my @n_list = @$n_ref;
  
  my %page_phrase_hash = %$phrase_hash_ref;
  if ($#n_list != 1) {
    trace("error: invalid n1, n2 list");
    return;
  }
  my $n1 = $n_list[0];
  my $n2 = $n_list[1];
  #$n1->dump();
  #$n2->dump();
  my @boundary_tag_list = ();
  #if immediate sibling, this means we have the record boundary already
  my $simple_boundary = 0;
  my $keyword_ele_index = -1;	# for example, Author, Title, Price. Title is the one that has the keyword, we remember its index in the record.
  #trace("before n1_right: ");
  my $n1_right = get_right_element($n1);
  #trace("n1_right: " . $n1_right->address);
  if ($n1_right->address() eq $n2->address()) {
    $simple_boundary = 1;
    $keyword_ele_index = 0;
    #trace("data record tag: " . $n1->tag() . "\n");
    push @boundary_tag_list, $n1->tag;
  }
  # this case has not been tested
  #$data_region->dump();
  if (!$simple_boundary) {
    my @content_list = $data_region->content_list;
    my $first;
    my $last;
    my $is_first = 1;
    #find first and last
    foreach my $ele (@content_list) {
      next if ((ref $ele) ne "HTML::Element");
      my $test_node = 
	$ele->look_down(
			sub {
			  exists($page_phrase_hash{trim($_[0]->as_text())} ) 
			});
      if (defined($test_node)) {
	$last = $ele;
	if ($is_first) {
	  $first = $ele;
	  $is_first = 0;
	}
      }
    }

    my $first_level1= get_level1_node($first, $data_region);
    my $last_level1= get_level1_node($last, $data_region);
    #trace_child_tags($data_region);
    my @before_tags = ();
    my @after_tags = ();

    my $left = $first_level1;
    my $right = $last_level1;
    my $n2_left = $n2;
    my $n1_right = $n1;

    #say we have a HHH-H*DD-H*DD-H*DD-KK pattern
    #from the first we get the left portion and from the last we get the right portion.
    while (1) {
      $left = $left->left;
      $n2_left = $n2_left->left;
      unless(is_element($left) && is_element($n2_left) 
	     && ($n2_left->address ne $n1->address)) {
	last;
      }
      if (has_same_structure($left, $n2_left)) {
	push @before_tags, $left->tag;
      } else {
	last;
      }
    }
    @before_tags = reverse @before_tags;
    while (1) {
      $right = $right->right;
      $n1_right = $n1_right->right;
      unless(is_element($right) && is_element($n1_right) && ($n2->address ne $n1_right->address)) {
	last;
      }
      if (has_same_structure($right, $n1_right)) {
	push @after_tags, $right->tag;
      } else {
	last;
      }
    }
    #trace "before_tags: $#before_tags:";
    #map {print $_} @before_tags;
    #trace("n1_tag: " . $n1->tag);
    #trace "after_tags: $#after_tags:";
    #map {print $_} @after_tags;
    push @boundary_tag_list, @before_tags;
    push @boundary_tag_list, $n1->tag;
    push @boundary_tag_list, @after_tags;
    $keyword_ele_index = $#before_tags + 1;
    #trace("boundary_tag_list in : $#boundary_tag_list");
  }				# !simple_boundary
  #trace("boundary_tag_list out: $#boundary_tag_list");
  return ($keyword_ele_index, \@boundary_tag_list);   
}

#we assume other pages having the same structure as the seed page
sub extract_other_page($\%$){
#input: * filepath
#       * boundary info list
#       * keyword
#output:* records_list 
#       * unmatched region addresses
#       * unmatched phrases
  my ($filepath, %boundary_info_list, $keyword) = @_;
  my $tree = get_tree($filepath);
  my @records_list = ();
  my @phrase_list = get_phrase_List_from_file($filepath);
  my %phrase_hash = get_phrase_hash(@phrase_list);
  my @unmatched_regions = ();
  while (my ($region_address, $boundary_info_ref) = 
	 each(%boundary_info_list)) {
    my @boundary_info = @$boundary_info_ref;
    my $region_tag = $boundary_info[0];
    my @boundary_tags = @{$boundary_info[1]};
    my $region = $tree->address($region_address);
    unless (defined($region) && ($region->tag eq $region_tag)) {
      push @unmatched_regions, $region->address;
      next;
    }
    my (@records, @matched_list) = 
      extract_region_by_boundary_info($region, $region_address, 
				      @boundary_info, %phrase_hash, $keyword);
    push @records_list, [@records];
    foreach my $phrase(@matched_list) {
      $phrase_hash{$phrase} = 0;
    }
  }
  my @unmatched_list = ();
  for my $phrase(@phrase_list) {
    push @unmatched_list, $phrase if ($phrase_hash{$phrase});
  }
  return (\@records_list, \@unmatched_regions, \@unmatched_list);
}
