package CCM::Util;
use Exporter;
our @ISA = ('Exporter');
our @EXPORT = qw(get_html_tree 
		 get_phrase_list_from_file 
		 get_phrase_list 
		 is_phrase_node 
		 get_phrase_node_list 
		 get_phrase_list_from_nodes
		 get_phrase_hash
		 get_level1_node
		 has_same_structure
		is_matching_leaf_node
		extract_phrases_from_site_pages
		get_site_pages
		get_seed_lineage
		compare_str_list
		get_right_element
		get_first_child_element
		is_element
		get_phrase_node_count
		get_phrase_node
		get_neighboring_phrase_node
		get_common_ancestor
		detach_records);

use strict;
use warnings;
use diagnostics;
use HTML::TreeBuilder;
use CCM::PerlUtil;
use CCM::Debug;

=head1 NAME

Utility

=head1 SYNOPSIS

General utility functions

=head1 DESCRIPTION

Functions that are not specific to data record extraction.

=cut

sub is_element($) {
  my $ele = shift;
  return (ref $ele eq "HTML::Element");
}
sub get_html_tree($){
  my $page_path = shift;
  my $proc_tree = HTML::TreeBuilder->new();
  return unless defined($proc_tree);
  $proc_tree->parse_file($page_path);
  return $proc_tree;
}

#we do not use regular expression because HTML::Element->as_text will
#convert & to &amp&, which will cause some mismatch.
#To simplify, we simply use HTML::Tree to get all the phrases.

sub get_phrase_list_from_file($$) {
  my ($file_path, $keyword) = @_;
  local *HTML_PAGE;
  open HTML_PAGE, $file_path or die "Fail to open $file_path: $!";
  my $sep_save = $/;
  undef $/;
  my $content = <HTML_PAGE>;
  my @real_phrase_list = get_phrase_list($content, $keyword);
  close HTML_PAGE;
  $/ = $sep_save;
  return @real_phrase_list;
}

sub get_refined_phrase_list(\@$){
  my ($phrase_list_ref, $keyword) = @_;
  my @phrase_list = @$phrase_list_ref;
  my @real_phrase_list = ();
  foreach my $phrase(@phrase_list) {
      $phrase = trim($phrase);
      if (lc($phrase) eq lc($keyword)) {
	next;
      }
      push @real_phrase_list, $phrase;
    }
  return @real_phrase_list;
}

sub get_phrase_list_from_nodes(\@$){
  my ($nodes_ref, $keyword) = @_;
  my @nodes = @$nodes_ref;
  my @phrase_list = map {$_->as_text;} @nodes;
  return get_refined_phrase_list(@phrase_list, $keyword);
}

sub get_phrase_list($$) {
  my ($content, $keyword) = @_;
  my $tree = HTML::TreeBuilder->new;
  $tree->parse($content);
  my @phrase_nodes = get_phrase_node_list($tree, $keyword);
  my @phrase_list = get_phrase_list_from_nodes(@phrase_nodes, $keyword);
  $tree->delete();
  return @phrase_list;
}

sub test_phrase_node($$&){
  my ($node, $to_match, $test) = @_;
  return 0 unless(defined($node));
  my @descendants = $node->descendants;
  if ($#descendants < 0) {
    my $text = $node->as_text;
#    if ($text =~ /Sun Certified Progr/) {
#      trace ('---' . $text . '---');
#      trace ('+++' . $to_match . '+++');
#    }
    return 1 if (&$test($text, $to_match));
  }
  return 0;
}
sub is_phrase_node($$){
  my ($node, $keyword) = @_;
  return test_phrase_node($node, $keyword, sub {$_[0] =~ $_[1]});
}

sub is_matching_leaf_node($$) {
  my ($node, $phrase) = @_;
  return test_phrase_node($node, $phrase, sub {trim($_[0]) eq trim($_[1])});
}

#find all the leaf node that has the keyword
sub get_phrase_node_list($$) {
  my ($tree, $keyword) = @_;
  #trace($tree->tag . " " . $keyword);
  my @node_list = $tree->look_down( 
				   sub {
				     return is_phrase_node($_[0], $keyword);}
				  );
  #trace_node_list(@node_list);
  return @node_list;
}

sub get_first_child_element{
  my $node = shift;
  my $idx = 0;
  my @lst = $node->content_list();
  my $child;
  while (1) {
    $child = $lst[$idx++];
    return $child if (is_element($child));
  }
}

sub get_right_element{
  my $node = shift;
  my $right = $node;
  my @lst = $node->parent->content_list();
  my $idx = $node->pindex;
  while (1) {
    #if we are the rightmost already, return undef
    return undef if ($idx == $#lst);
    $right = $lst[++$idx];
    return $right unless (defined($right));
    return $right if ((ref $right) eq 'HTML::Element');
  }
}
sub get_level1_node($$){
  my ($node, $region) = @_;

  my $temp_node = $node;
  my $level1_node;
  #trace("get_level1_node: " . $region->tag);
  while (defined($temp_node->parent) && ($temp_node->address ne $region->address)) {
      $level1_node = $temp_node;
      #trace("get_level1_node: " . $level1_node->tag);
      $temp_node = $temp_node->parent;
    }
  return $level1_node;
}

sub get_phrase_hash{
  my @phrase_list = @_;
  my %phrase_hash = ();
  for my $phrase (@phrase_list) {
    if (exists($phrase_hash{$phrase})) {
      ++$phrase_hash{$phrase};
    } else {
      $phrase_hash{$phrase} = 1 ;
    }
  }
  return %phrase_hash;
}

sub has_same_structure {
  my ($ele1, $ele2) = @_;
  my @descendent_tags1 = map {$_->tag} $ele1->descendents; 
  my @descendent_tags2 = map {$_->tag} $ele2->descendents; 
  return (@descendent_tags1 == @descendent_tags2);
}

#right now simply match number of descendent elements, 
# but can make it more accurate if needed.
sub match_node_with_boundary{
  my ($ele1, $count) = @_;
  my @descendent_tags1 = map {$_->tag} $ele1->descendents; 
  return (@descendent_tags1 == $count);
}
sub match_nodes_with_boundary_info{
  my ($nodes_ref, $boundary_info_ref) = @_;

}

sub get_site_pages($){
  my $datadir = shift;
  my %site_pages = (); # key is the site, value is a list of all pages
  print "CCM start processing: $datadir\n";
  opendir (DATAPATH, $datadir) or die "Fail to open $datadir: $!";
  while (defined(my $fsname = readdir (DATAPATH))) {
    my $full_fsname = "$datadir\\$fsname";
    if ((-d $full_fsname) && ($fsname ne '.') && ($fsname ne '..')) {
      my @pages = ();
      opendir (SITEPATH, $full_fsname) or die "Fail to open $full_fsname: $!";
      while (defined (my $inner_fsname = readdir(SITEPATH))) {
	my $inner_full_fsname = "$full_fsname\\$inner_fsname";
	if ((-f $inner_full_fsname) && ($inner_fsname =~ /\.htm(|l)/)) {
	  #print "$inner_full_fsname\n";
	  push @pages, $inner_fsname;
	}
      }
      $site_pages{$fsname} = [@pages];
      closedir(SITEPATH);
    }
  }
  closedir(DATAPATH);
  return %site_pages;
}
#use HTML::TreeBuilder to extract instead of regular expression to
#keep special character consistent like &amp;
sub extract_phrases_from_site_pages($$%){
  my $datadir = shift;
  my $keyword = shift;
  my %site_pages = @_;
  my %ranked_phrase_list = ();
  my %page_phrase_list = (); # hash of hash: site -> (page -> phrase)
  my $sep_save = $/;
  undef $/;
  while ((my $each_site, my $pages) = each(%site_pages)) {
    #print "$each_site:\n";
    foreach my $each_page (@{$pages}) {
      my $file_path = "$datadir\\$each_site\\$each_page";
      open HTML_PAGE, $file_path or die "Fail to open $file_path: $!";
      my $content = <HTML_PAGE>;
#      my @phrase_list = ($content =~ /\>\s*([^><]*$keyword[^><]*)\s*\</ig);
      my @real_phrase_list = get_phrase_list($content, $keyword);
      foreach my $phrase (@real_phrase_list) {
	if (exists($ranked_phrase_list{$phrase})) {
	  ++$ranked_phrase_list{$phrase};
	} else {
	  $ranked_phrase_list{$phrase} = 1;
	}
      }
      $page_phrase_list{$each_site}{$each_page} = @real_phrase_list;
      close HTML_PAGE;
    }
  }
  $/ = $sep_save;
  return (\%ranked_phrase_list, \%page_phrase_list);
}

sub get_seed_relative_address($$){
  my ($seed, $level1) = @_;
  my $seed_addr = $seed->address;
  my $level1_addr = $level1->address;
  my $relative_addr = substr($seed_addr, (length $level1_addr) + 1);
  return $relative_addr;
}

sub get_seed_lineage($$){
  my ($seed, $level1) = @_;
  my @seed_lineage_to_root = $seed->lineage;
  my @level1_lineage_to_root = $level1->lineage;
  my @seed_lineage = 
    splice (@seed_lineage_to_root, 
      $#seed_lineage_to_root - $#level1_lineage_to_root);
  my $seed_relative_addr = get_seed_relative_address($seed, $level1);
  return ($seed_relative_addr, $seed->tag, (map {$_->tag} @seed_lineage_to_root));
}

sub compare_str_list(\@\@) {
  my ($list1_ref, $list2_ref) = @_;
  my @list1 = @$list1_ref;
  my @list2 = @$list2_ref;
  return 0 if ($#list1 != $#list2);
  for my $i (0..$#list1) {
    return 0 if ($list1[$i] ne $list2[$i]);
  }
  return 1;
}

sub get_common_ancestor($$) {
  #trace("function: get_common_ancestor");
  my ($node1, $node2) = @_;
  my $address1 = $node1->address;
  my $address2 = $node2->address;
  #trace ("address1:" . $address1);
  #trace ("address2:" . $address2);
  my @list1 = split (/\./, $address1);
  my @list2 = split (/\./, $address2);
  my $count = ($#list1 < $#list2) ? $#list1 : $#list2;
  my $address = "";
  my $i = 0;
  while ($list1[$i] eq $list2[$i] && ($i <= $count)) {
    $address .= $list1[$i++] . '.';
  }
  #$address = substr($address, 0, (length $address) - 1);
  #trace ("ancestor address:" . $address);
  return $node1->address($address);
}

sub get_phrase_node($$) {
  my ($tree, $keyword) = @_;
  my $phrase_node = 
    $tree->look_down(sub {is_matching_leaf_node($_[0],$keyword)});
  return $phrase_node;
} 

#sometimes we have several data records having the same phrase, we need
#to make sure to find different node when finding neighboring node.
sub get_neighboring_phrase_node($$$) {
  my ($tree, $keyword, $node) = @_;
  my $phrase_node = 
    $tree->look_down(
		     sub {defined($_[0]) && is_element($_[0]) 
			    && is_matching_leaf_node($_[0],$keyword) 
			      && ($node->address ne $_[0]->address)});
  return $phrase_node;
} 

sub get_phrase_node_count($$) {
  my ($tree, $keyword) = @_;
  my @node_list = get_phrase_node_list($tree, $keyword);
  return ($#node_list + 1);
}

sub detach_records {
  my @record_list = @_;
  my $record_count = 1;
  for my $each_records(@record_list) {
    my @records = @$each_records;
    map {$_->detach} @records;
  }
}
